/*
 * TriCore cpu parameters for qemu.
 *
 * Copyright (c) 2012-2014 Bastian Koppelmann C-Lab/University Paderborn
 * SPDX-License-Identifier: LGPL-2.1+
 */

#ifndef TRICORE_CPU_PARAM_H
#define TRICORE_CPU_PARAM_H 1

#define TARGET_LONG_BITS 32
#define TARGET_PAGE_BITS 14
#define TARGET_PHYS_ADDR_SPACE_BITS 32
#define TARGET_VIRT_ADDR_SPACE_BITS 32
#define NB_MMU_MODES 3

#endif
