/* Automatically generated by create_config - do not modify */
#define HOST_I386 1
#define CONFIG_WIN32 1
#define CONFIG_TCG 1
#define CONFIG_CPUID_H 1
// #define CONFIG_INT128 1
#define CONFIG_CMPXCHG128 1
// #define CONFIG_ATOMIC64 1
#define CONFIG_PLUGIN 1
